#include <stdlib.h>
#include "grafos.h"

struct dijNo
{
    int    indice;
    int    indiceNoAnterior;
    double distancia;
    struct dijNo* proximoNoLista;
};

void dijkstra(int nNos, int nArcos, const int* no1, const int* no2, const double* compArco, int noInicio, double* distNo, int* noAnterior)
{
    /*----------------------------------------------------------------------------
    Algoritmo de Dijkstra, para encontrar o caminho mnimo em um grafo.

    Dados de entrada:
    nNos:       Numero de nos
    nArcos:     Numero de arcos
    no1         Vetor de tamanho [nArcos] indicando o no' inicial de cada arco
    no2         Vetor de tamanho [nArcos] indicando o no' final de cada arco
    compArco    Vetor de tamanho [nArcos] indicando o comprimento de cada arco
    noInicio    No inicial do grafo

    Dados de saida:
    distNo      Vetor de tamanho [nNos] indicando a distancia do no' ao n inicial
    noAnterior  Vetor de tamanho [nNos] indicando o no' anterior no caminho minimo

    ----------------------------------------------------------------------------*/
    struct dijNo* noInicialLista = malloc(sizeof(struct dijNo));
    struct dijNo* noAux1;
    struct dijNo* noEntrando;
    double distancia;
    int idNoAnterior;
    int indice;
    int i;

    /* Inicializa vetores de saida com -1 */
    for(i = 0; i < nNos; i++)
    {
        distNo[i] = -1;
        noAnterior[i] = -1;
    }
    /* Inicia primeiro no' */
    noInicialLista->indice = noInicio;
    noInicialLista->indiceNoAnterior = noInicio;
    noInicialLista->distancia = 0;
    while (noInicialLista != NULL)
    {
	/* Guarda dados do no */
        distancia = noInicialLista->distancia;
        indice = noInicialLista->indice;
        idNoAnterior = noInicialLista->indiceNoAnterior;
	/* Remove primeiro no da lista */
        noAux1 = noInicialLista->proximoNoLista;
        free(noInicialLista);
        noInicialLista = noAux1;
	/* Se o no nao tiver sido incluido ainda, */
        if (noAnterior[indice] == -1)
        {
	    /* adiciona-lo ao mapa de distancias dos nos */
            distNo[indice] = distancia;
            noAnterior[indice] = idNoAnterior;
	    /* Colocar todos os nos que este no' alcanca na lista */
            for(i = 0; i < nArcos; i++)
            {
                if(indice == no1[i])
                {
		    /* Inserir no' na lista em ordem crescente de distancia */
                    noEntrando = malloc(sizeof(struct dijNo));
                    noEntrando->indice = no2[i];
                    noEntrando->indiceNoAnterior = indice;
                    noEntrando->distancia = distancia + compArco[i];
                    if(noInicialLista == NULL)
                    {
                        noInicialLista = noEntrando;
                        noEntrando->proximoNoLista = NULL;
                    }
                    else
                    {
                        if(noInicialLista->distancia > noEntrando->distancia)
                        {
                            noEntrando->proximoNoLista = noInicialLista;
                            noInicialLista = noEntrando;
                        }
                        else
                        {
                            noAux1 = noInicialLista;
                            while((noAux1->proximoNoLista != NULL) && (noAux1->proximoNoLista->distancia < noEntrando->distancia))
                            {
                                noAux1 = noAux1->proximoNoLista;
                            }
                            noEntrando->proximoNoLista = noAux1->proximoNoLista;
                            noAux1->proximoNoLista = noEntrando;
                        }
                    }
                }
            }
        }
    }
}
